/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001,2002 Nokia Corporation
   Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   Original Author: Guruprasad Krishnamurthy <kgprasad@hotmail.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   */

/*
   $Id: sdp.h,v 1.43 2003/02/28 17:31:29 kds Exp $

   Top level SDP definitions for attribute types. 
   This file provides a "one-stop" change for any
   modifications to the attributes

   Fixes:
		Dmitry Kasatkin		: bug fixes, cleanup, re-arrangement, 
					  continuation state, mtu
*/

#ifndef SDP_H
#define SDP_H

#include <stdint.h>

#include <affix/bluetooth.h>
#include <affix/utils.h>

__BEGIN_DECLS


/*
 ** All definitions are based on Bluetooth Assigned Numbers
 ** of the Bluetooth Specification
 **
 **
 ** NOTE : The quotes("") are needed as this value is treated
 ** as a string in uuidHandler.c
 **
 ** NOTE : The conversion of this string to 128bit relies
 ** in preserving the sequence shown below ie. if you need
 ** to change, then preserve the "-" characters in the 
 ** appropriate places. If their positions CHANGE, then you need to
 **
 */
#define BASE_UUID "00000000-0000-1000-8000-00805F9B34FB"

/*
 ** Protocol UUIDs
 */
#define SDP_UUID	0x0001
#define UDP_UUID	0x0002
#define RFCOMM_UUID	0x0003
#define TCP_UUID	0x0004
#define TCS_BIN_UUID	0x0005
#define TCS_AT_UUID	0x0006
#define OBEX_UUID	0x0008
#define IP_UUID		0x0009
#define FTP_UUID        0x000A
#define HTTP_UUID	0x000C
#define WSP_UUID	0x000E
#define BNEP_UUID	0x000F
#define L2CAP_UUID	0x0100

/*
 ** Service class identifiers of standard services and
 ** service groups
 */
#define SDPServerServiceClassID      		0x1000
#define BrowseGroupDescServiceClassID 		0x1001
#define PublicBrowseGroup                       0x1002
#define SerialPortServiceClassID     		0x1101
#define LANAccessUsingPPPServiceClassID     	0x1102
#define DialupNetworkingServiceClassID     	0x1103
#define IrMCSyncServiceClassID     		0x1104
#define ObexObjectPushServiceClassID     	0x1105
#define ObexFileTransferServiceClassID     	0x1106
#define IrMCSyncCommandServiceClassID     	0x1107
#define HeadsetServiceClassID     		0x1108
#define CordlessTelephonyServiceClassID  	0x1109
#define IntercomServiceClassID  		0x1110
#define FaxServiceClassID  			0x1111
#define HeadsetAudioGatewayServiceClassID  	0x1112
#define PANUServiceClassID			0x1115
#define NAPServiceClassID			0x1116
#define GNServiceClassID			0x1117
#define HandsFreeServiceClassID			0x111E
#define AGHandsFreeServiceClassID		0x111F
#define PnPInformationServiceClassID  		0x1200
#define GenericNetworkingServiceClassID  	0x1201
#define GenericFileTransferServiceClassID  	0x1202
#define GenericAudioServiceClassID  		0x1203
#define GenericTelephonyServiceClassID  	0x1204

/*
 ** Attribute identifier codes
 */
#define SDPServerServiceRecordHandle 	0x0000

/*
 ** Possible values for uint16_t are
 ** listed below. See SDP Spec, section "Service Attribute
 ** Definitions" for more details.
 */
#define AttrServiceRecordHandle         0x0000
#define AttrServiceClassIDList          0x0001
#define AttrServiceRecordState          0x0002
#define AttrServiceID                   0x0003
#define AttrProtocolDescriptorList      0x0004
#define AttrBrowseGroupList             0x0005
#define AttrLanguageBasedAttrIDList     0x0006
#define AttrServiceInfoTimeToLive       0x0007
#define AttrServiceAvailability         0x0008
#define AttrProfileDescriptorList	0x0009
#define AttrDocumentationURL            0x000A
#define AttrClientExecutableURL         0x000B
/* HandsFree */
#define AttrSupportedFeatures		0x0311

/* PAN */
#define AttrIpSubnet			0x0200
#define AttrSecurityDescription		0x030A
#define AttrNetAccessType		0x030B
#define AttrMaxNetAccessRate		0x030C

/*
 ** There is a conflict with the SDP spec and
 ** assigned numbers doc in which this attribute
 ** has a value of 0x000D
 */
#define AttrIconURL                     0x000C                      

/*
 ** These identifiers are based on the SDP
 ** spec stating that "base attribute id of
 ** the primary (universal) language must be 0x0100
 */
#define PrimaryLangBase 		0x0100

#define AttrServiceName_Primary         	0x0000+PrimaryLangBase
#define AttrServiceDescription_Primary  	0x0001+PrimaryLangBase
#define AttrServiceProviderName_Primary 	0x0002+PrimaryLangBase
/*
 ** Other languages should have theor own offset ex.
 ** #define XXXLangBase yyyy
 ** #define AttrServiceName_XXX	0x0000+XXXLangBase
 ** ..
 */

/*
 ** These attributes are specific to the SDP server only
 ** ie. can be present only in the sdpsvc_t of the
 ** SDP server
 */
#define AttrVersionNumberList           0x0200
#define AttrServiceDatabaseState        0x0201           

#define AttrGroupID			0x0200

/* --------------------------------------------------------------- */

/*
 ** The Data representation in SDP PDUs (pg 339 and 340 of BT SDP Spec)
 **
 ** These are the exact data type+size descriptor values
 ** that go into PDU buffer.
 **
 ** The datatype(leading 5bits) + size descriptor(last 3 bits)
 ** is 8 bits. The size descriptor is critical to extract the
 ** right number of bytes for the data value from the PDU.
 **
 ** For most basic types, the datatype+size descriptor is
 ** straightforward. However for constructed types and strings,
 ** the size of the data is in the next "n" bytes following the
 ** 8 bits (datatype+size) descriptor. Exactly what the "n" is
 ** specified in the 3 bits of the data size descriptor.
 **
 ** TextString and URLString can be of size 2^{8, 16, 32} bytes
 ** DataSequence and DataSequenceAlternates can be of size 2^{8, 16, 32}
 ** The size are computed post-facto in the API and are not known apriori
 */
/*
 ** Permitted values for DTD
 */
#define Data_Nil 			0x00
#define UnsignedInteger_8  		0x08
#define UnsignedInteger_16		0x09
#define UnsignedInteger_32		0x0A
#define UnsignedInteger_64		0x0B
#define UnsignedInteger_128		0x0C
#define Integer_8			0x10
#define Integer_16			0x11
#define Integer_32			0x12
#define Integer_64			0x13
#define Integer_128			0x14
#define UUID_XX				0x18
#define UUID_16				0x19
#define UUID_32				0x1A
#define UUID_128			0x1C
#define TextStringSize8			0x25
#define TextStringSize16		0x26
#define TextStringSize32        	0x27
#define Boolean				0x28
#define SequenceSize8			0x35
#define SequenceSize16			0x36
#define SequenceSize32			0x37
#define AlternativeSize8		0x3D
#define AlternativeSize16		0x3E
#define AlternativeSize32		0x3F
#define URLStringSize8			0x45
#define URLStringSize16			0x46
#define URLStringSize32			0x47

#define SDP_DTD_NULL			0x00
#define SDP_DTD_UINT			0x01
#define SDP_DTD_INT			0x02
#define SDP_DTD_UUID			0x03
#define SDP_DTD_STRING			0x04
#define SDP_DTD_BOOL			0x05
#define SDP_DTD_SEQ			0x06
#define SDP_DTD_ALT			0x07
#define SDP_DTD_URL			0x08

#define SDP_DTD_8			0x00
#define SDP_DTD_16			0x01
#define SDP_DTD_32			0x02
#define SDP_DTD_64			0x03
#define SDP_DTD_128			0x04
#define SDP_DTD_SIZE8			0x05
#define SDP_DTD_SIZE16			0x06
#define SDP_DTD_SIZE32			0x07

/*
 ** The PDU identifiers of SDP packets between client
 ** and server
 */
#define SDP_ErrorResponse			0x01
#define SDP_ServiceSearchRequest 		0x02
#define SDP_ServiceSearchResponse 		0x03
#define SDP_ServiceAttributeRequest 		0x04
#define SDP_ServiceAttributeResponse 		0x05
#define SDP_ServiceSearchAttributeRequest 	0x06
#define SDP_ServiceSearchAttributeResponse 	0x07


/*
 ** Some additions to support service registration.
 ** These are outside the scope of the Bluetooth specification
 */
#define SDP_ServiceRegistrationRequest 		0x75
#define SDP_ServiceRegistrationResponse 	0x76
#define SDP_ServiceUpdateRequest		0x77
#define SDP_ServiceUpdateResponse		0x78
#define SDP_ServiceRemovalRequest		0x79
#define SDP_ServiceRemovalResponse		0x80


/*
 ** SDP Error codes
 */
#define SDP_ERR_SDP_VERSION		0x0001
#define SDP_ERR_SERVICE_RECORD_HANDLE	0x0002
#define SDP_ERR_REQUEST_SYNTAX		0x0003
#define SDP_ERR_PDU_SIZE		0x0004
#define SDP_ERR_CONT_STATE		0x0005
#define SDP_ERR_RESOURCES		0x0006
/* internal Affix specific */
#define SDP_ERR_INVALID_ARG		0x0101
#define SDP_ERR_NOT_EXIST		0x0102
#define SDP_ERR_SYNTAX			0x0103
#define SDP_ERR_INTERNAL		0x0104
#define SDP_ERR_SERVER			0x0105


#define RequestBufferSize		2048
#define ResponseBufferSize		USHRT_MAX

#define BasicAttributePDUFormSize 	32
#define SequencePDUFormSize 	  	128
#define UUIDSequenceSize	  	256


#define SDP_TIMEOUT	20

typedef struct {
	char data[16];
} uint128_t ;

typedef uint128_t	int128_t;

typedef struct {
	int	type;
	union {
		uint16_t	uuid16Bit;
		uint32_t	uuid32Bit;
		uint128_t	uuid128Bit;
	} value ;
} uuid_t;      


/*
 ** User visible strings can be in many languages
 ** in addition to the universal language.
 **
 ** Language meta-data includes language code in
 ** ISO639 followed by the encoding format.
 ** The third field in this structure is the
 ** attribute offset for the language. User visible
 ** strings in the specified language can be obtained
 ** at this offset
 */
typedef struct {
	uint16_t	lang;
	uint16_t	encoding;
	uint16_t	offset;
} sdp_lang_t ;

/* ------------------------------------------------------- */




typedef enum operationMode {
	SDP_CLIENT,
	SDP_SERVER
} sdp_mode_t;

/*
 ** Request blob as received from the client
 */
typedef struct {
	int	fd;
	int	mtu;
	char	*data;
} sdp_request_t ;

/*
 ** Proprietary (Nokia) extensions to register/delete/modify
 ** service records
 */
typedef struct {
	uint8_t		pduId;
	uint16_t	transactionId;
	uint16_t 	paramLength;	/* not includeing pduId and transId */
	uint16_t	data[0];
} __PACK__ sdp_hdr_t;

struct sdp_error {
	uint8_t		pduId;
	uint16_t	transactionId;
	uint16_t 	paramLength;	/* not includeing pduId and transId */
	uint16_t	ErrorCode;
	uint8_t		ErrorInfo[0];
} __PACK__;

/*
 ** The service record when encapsulated as a PDU
 ** as per BT spec
 **
 **  Each of the PDU form objects are cached to
 ** avoid repetitive encodings on the server side
 ** They are however set to null on the client side
 */
typedef struct {
	char	*data;
	int	length;
	int	size;
} sdppdu_t;     


/* ---------------------------------------------------- */

/* SDP record state */
enum {
	SDP_STATE_CREATED,
	SDP_STATE_REGISTERED,
	SDP_STATE_UPDATE_READY,
	SDP_STATE_REGISTER_READY,
	SDP_STATE_NOT_EXIST
};

/*
 ** sdpsvc_t is a composite of many attributes not
 ** all of which need be present. References of those
 ** composites not present would be set to NULL
 **
 ** A set of mandatory attributes will be defined and
 ** when set, the sdpsvc_t gets into a "commit ready"
 ** state
 **
 */
typedef struct {
	uint32_t	serviceRecordHandle;
	BD_ADDR		*bda;		/* service owner */
	int		state;
	sdppdu_t	pdu;	/* PDU form and the entire service record */
	slist_t 	*targetPattern;	/* all possible uuid_t */
	slist_t 	*attributeList;
	int		fd;	/* to remove service */

} sdpsvc_t ;

/* -------------------------------------- */

#define SDP_DTD_TYPE(dtd)		((dtd)>>3)
#define SDP_DTD_SIZE(dtd)		((dtd)&0x07)
#define SDP_DTD_SETSIZE(dtd, size)	((dtd) = ((dtd)&~0x07) | size)

typedef struct sdpdata sdpdata_t;
struct sdpdata {
	uint8_t		dtd;
	uint16_t	attrId;
	union {
		int8_t 		int8;
		int16_t 	int16;
		int32_t 	int32;
		int64_t 	int64;
		uint128_t 	int128;
		uint8_t 	uint8;
		uint16_t 	uint16;
		uint32_t 	uint32;
		uint64_t 	uint64;
		uint128_t 	uint128;
		uuid_t		uuid;
		char 		*stringPtr;
		slist_t	 	*dataSeq;
	} value;
	sdppdu_t 	pdu;
	int 		unitSize;
};

sdpsvc_t *sdp_create_svc(void);
void sdp_free_svc(sdpsvc_t *svcRec);
void sdp_free_svclist(slist_t **svcList);
void sdp_free_data(sdpdata_t *pData);
void sdp_free_seq(slist_t **pDataSeq);
sdpdata_t *sdp_get_attr(sdpsvc_t *svcRec, uint16_t attrId);
sdpdata_t *sdp_create_data(uint8_t dtd, void *pValue);

static inline int sdp_is_seq(sdpdata_t *data)
{
	return (SDP_DTD_TYPE(data->dtd) == SDP_DTD_SEQ);
}

static inline int sdp_is_alt(sdpdata_t *data)
{
	return (SDP_DTD_TYPE(data->dtd) == SDP_DTD_ALT);
}

static inline int sdp_is_seqalt(sdpdata_t *data)
{
	return (sdp_is_seq(data) || sdp_is_alt(data));
}

char *sdp_error(int err);

__END_DECLS

#endif //SDP_H
